From c970a38889d5774f7d0f73215dfd70934524d76a Mon Sep 17 00:00:00 2001
From: Boris Kolpackov <boris@codesynthesis.com>
Date: Tue, 10 Jul 2012 15:17:13 +0200
Subject: Add support for custom database type mapping

New pragma qualifier, map, and specifiers: as, to, from. New tests:
<database>/custom.
---
 odb/mysql/details/conversion.hxx | 59 ++++++++++++++++++++++++++++++++++++++
 odb/mysql/query.cxx              | 18 +++++++++++-
 odb/mysql/query.hxx              | 61 +++++++++++++++++++++++++---------------
 odb/mysql/query.ixx              | 10 ++++---
 odb/mysql/query.txx              | 32 ++++++++++-----------
 5 files changed, 136 insertions(+), 44 deletions(-)
 create mode 100644 odb/mysql/details/conversion.hxx

diff --git a/odb/mysql/details/conversion.hxx b/odb/mysql/details/conversion.hxx
new file mode 100644
index 0000000..4d473f8
--- /dev/null
+++ b/odb/mysql/details/conversion.hxx
@@ -0,0 +1,59 @@
+// file      : odb/mysql/details/conversion.hxx
+// copyright : Copyright (c) 2009-2012 Code Synthesis Tools CC
+// license   : GNU GPL v2; see accompanying LICENSE file
+
+#ifndef ODB_MYSQL_DETAILS_CONVERSION_HXX
+#define ODB_MYSQL_DETAILS_CONVERSION_HXX
+
+#include <odb/mysql/traits.hxx>
+
+#include <odb/details/meta/answer.hxx>
+
+namespace odb
+{
+  // @@ Revise this.
+  //
+  namespace details {}
+
+  namespace mysql
+  {
+    namespace details
+    {
+      using namespace odb::details;
+
+      // Detect whether conversion is specified in type_traits.
+      //
+      template <typename T>
+      meta::yes
+      conversion_p_test (typename type_traits<T>::conversion*);
+
+      template <typename T>
+      meta::no
+      conversion_p_test (...);
+
+      template <typename T>
+      struct conversion_p
+      {
+        static const bool value =
+          sizeof (conversion_p_test<T> (0)) == sizeof (meta::yes);
+      };
+
+      template <typename T, bool = conversion_p<T>::value>
+      struct conversion;
+
+      template <typename T>
+      struct conversion<T, true>
+      {
+        static const char* to () {return type_traits<T>::conversion::to ();}
+      };
+
+      template <typename T>
+      struct conversion<T, false>
+      {
+        static const char* to () {return 0;}
+      };
+    }
+  }
+}
+
+#endif // ODB_MYSQL_DETAILS_CONVERSION_HXX
diff --git a/odb/mysql/query.cxx b/odb/mysql/query.cxx
index 2573ac1..70d55c2 100644
--- a/odb/mysql/query.cxx
+++ b/odb/mysql/query.cxx
@@ -118,10 +118,13 @@ namespace odb
     }
 
     void query::
-    add (details::shared_ptr<query_param> p)
+    add (details::shared_ptr<query_param> p, const char* conv)
     {
       clause_.push_back (clause_part (clause_part::param));
 
+      if (conv != 0)
+        clause_.back ().part = conv;
+
       parameters_.push_back (p);
       bind_.push_back (MYSQL_BIND ());
       binding_.bind = &bind_[0];
@@ -258,7 +261,20 @@ namespace odb
             if (last != ' ' && last != '(')
               r += ' ';
 
+            // Add the conversion expression, if any.
+            //
+            string::size_type p;
+            if (!i->part.empty ())
+            {
+              p = i->part.find ("(?)");
+              r.append (i->part, 0, p);
+            }
+
             r += '?';
+
+            if (!i->part.empty ())
+              r.append (i->part, p + 3, string::npos);
+
             break;
           }
         case clause_part::native:
diff --git a/odb/mysql/query.ixx b/odb/mysql/query.ixx
index 674c6cd..2508ae3 100644
--- a/odb/mysql/query.ixx
+++ b/odb/mysql/query.ixx
@@ -8,20 +8,22 @@ namespace odb
   {
     template <typename T, database_type_id ID>
     inline void query::
-    append (val_bind<T> v)
+    append (val_bind<T> v, const char* conv)
     {
       add (
         details::shared_ptr<query_param> (
-          new (details::shared) query_param_impl<T, ID> (v)));
+          new (details::shared) query_param_impl<T, ID> (v)),
+        conv);
     }
 
     template <typename T, database_type_id ID>
     inline void query::
-    append (ref_bind<T> r)
+    append (ref_bind<T> r, const char* conv)
     {
       add (
         details::shared_ptr<query_param> (
-          new (details::shared) query_param_impl<T, ID> (r)));
+          new (details::shared) query_param_impl<T, ID> (r)),
+        conv);
     }
   }
 }
diff --git a/odb/mysql/query.txx b/odb/mysql/query.txx
index 0377c33..80b6358 100644
--- a/odb/mysql/query.txx
+++ b/odb/mysql/query.txx
@@ -19,7 +19,7 @@ namespace odb
       //
       append (c.table (), c.column ());
       append ("=");
-      append<bool, ID> (val_bind<bool> (true));
+      append<bool, ID> (val_bind<bool> (true), c.conversion ());
     }
 
     // query_column
@@ -30,9 +30,9 @@ namespace odb
     {
       query q (table_, column_);
       q += "IN (";
-      q.append<T, ID> (val_bind<T> (v1));
+      q.append<T, ID> (val_bind<T> (v1), conversion_);
       q += ",";
-      q.append<T, ID> (val_bind<T> (v2));
+      q.append<T, ID> (val_bind<T> (v2), conversion_);
       q += ")";
       return q;
     }
@@ -43,11 +43,11 @@ namespace odb
     {
       query q (table_, column_);
       q += "IN (";
-      q.append<T, ID> (val_bind<T> (v1));
+      q.append<T, ID> (val_bind<T> (v1), conversion_);
       q += ",";
-      q.append<T, ID> (val_bind<T> (v2));
+      q.append<T, ID> (val_bind<T> (v2), conversion_);
       q += ",";
-      q.append<T, ID> (val_bind<T> (v3));
+      q.append<T, ID> (val_bind<T> (v3), conversion_);
       q += ")";
       return q;
     }
@@ -58,13 +58,13 @@ namespace odb
     {
       query q (table_, column_);
       q += "IN (";
-      q.append<T, ID> (val_bind<T> (v1));
+      q.append<T, ID> (val_bind<T> (v1), conversion_);
       q += ",";
-      q.append<T, ID> (val_bind<T> (v2));
+      q.append<T, ID> (val_bind<T> (v2), conversion_);
       q += ",";
-      q.append<T, ID> (val_bind<T> (v3));
+      q.append<T, ID> (val_bind<T> (v3), conversion_);
       q += ",";
-      q.append<T, ID> (val_bind<T> (v4));
+      q.append<T, ID> (val_bind<T> (v4), conversion_);
       q += ")";
       return q;
     }
@@ -75,15 +75,15 @@ namespace odb
     {
       query q (table_, column_);
       q += "IN (";
-      q.append<T, ID> (val_bind<T> (v1));
+      q.append<T, ID> (val_bind<T> (v1), conversion_);
       q += ",";
-      q.append<T, ID> (val_bind<T> (v2));
+      q.append<T, ID> (val_bind<T> (v2), conversion_);
       q += ",";
-      q.append<T, ID> (val_bind<T> (v3));
+      q.append<T, ID> (val_bind<T> (v3), conversion_);
       q += ",";
-      q.append<T, ID> (val_bind<T> (v4));
+      q.append<T, ID> (val_bind<T> (v4), conversion_);
       q += ",";
-      q.append<T, ID> (val_bind<T> (v5));
+      q.append<T, ID> (val_bind<T> (v5), conversion_);
       q += ")";
       return q;
     }
@@ -101,7 +101,7 @@ namespace odb
         if (i != begin)
           q += ",";
 
-        q.append<T, ID> (val_bind<T> (*i));
+        q.append<T, ID> (val_bind<T> (*i), conversion_);
       }
       q += ")";
       return q;
diff --git a/odb/mysql/query.hxx b/odb/mysql/query.hxx
index c0f812a..732244c 100644
--- a/odb/mysql/query.hxx
+++ b/odb/mysql/query.hxx
@@ -24,6 +24,7 @@
 #include <odb/details/shared-ptr.hxx>
 
 #include <odb/mysql/details/export.hxx>
+#include <odb/mysql/details/conversion.hxx>
 
 namespace odb
 {
@@ -99,7 +100,7 @@ namespace odb
         clause_part (bool p): kind (boolean), bool_part (p) {}
 
         kind_type kind;
-        std::string part;
+        std::string part; // If kind is param, then part is conversion expr.
         bool bool_part;
       };
 
@@ -142,7 +143,8 @@ namespace odb
       query (val_bind<T> v)
         : binding_ (0, 0)
       {
-        append<T, type_traits<T>::db_type_id> (v);
+        append<T, type_traits<T>::db_type_id> (
+          v, details::conversion<T>::to ());
       }
 
       template <typename T>
@@ -150,7 +152,8 @@ namespace odb
       query (ref_bind<T> r)
         : binding_ (0, 0)
       {
-        append<T, type_traits<T>::db_type_id> (r);
+        append<T, type_traits<T>::db_type_id> (
+          r, details::conversion<T>::to ());
       }
 
       template <database_type_id ID>
@@ -221,7 +224,8 @@ namespace odb
       query&
       operator+= (val_bind<T> v)
       {
-        append<T, type_traits<T>::db_type_id> (v);
+        append<T, type_traits<T>::db_type_id> (
+          v, details::conversion<T>::to ());
         return *this;
       }
 
@@ -229,18 +233,19 @@ namespace odb
       query&
       operator+= (ref_bind<T> r)
       {
-        append<T, type_traits<T>::db_type_id> (r);
+        append<T, type_traits<T>::db_type_id> (
+          r, details::conversion<T>::to ());
         return *this;
       }
 
     public:
       template <typename T, database_type_id ID>
       void
-      append (val_bind<T>);
+      append (val_bind<T>, const char* conv);
 
       template <typename T, database_type_id ID>
       void
-      append (ref_bind<T>);
+      append (ref_bind<T>, const char* conv);
 
       void
       append (const std::string& native);
@@ -250,7 +255,7 @@ namespace odb
 
     private:
       void
-      add (details::shared_ptr<query_param>);
+      add (details::shared_ptr<query_param>, const char* conv);
 
     private:
       typedef std::vector<clause_part> clause_type;
@@ -391,10 +396,11 @@ namespace odb
     template <typename T, database_type_id ID>
     struct query_column
     {
       // Note that we keep shalow copies of the table and column names.
       //
-      query_column (const char* table, const char* column)
-          : table_ (table), column_ (column)
+      query_column (const char* table, const char* column, const char* conv)
+          : table_ (table), column_ (column), conversion_ (conv)
       {
       }
 
@@ -410,6 +416,14 @@ namespace odb
         return column_;
       }
 
+      // Can be NULL.
+      //
+      const char*
+      conversion () const
+      {
+        return conversion_;
+      }
+
       // is_null, is_not_null
       //
     public:
@@ -462,7 +476,7 @@ namespace odb
       {
         query q (table_, column_);
         q += "=";
-        q.append<T, ID> (v);
+        q.append<T, ID> (v, conversion_);
         return q;
       }
 
@@ -479,7 +493,7 @@ namespace odb
       {
         query q (table_, column_);
         q += "=";
-        q.append<T, ID> (r);
+        q.append<T, ID> (r, conversion_);
         return q;
       }
 
@@ -547,7 +561,7 @@ namespace odb
       {
         query q (table_, column_);
         q += "!=";
-        q.append<T, ID> (v);
+        q.append<T, ID> (v, conversion_);
         return q;
       }
 
@@ -564,7 +578,7 @@ namespace odb
       {
         query q (table_, column_);
         q += "!=";
-        q.append<T, ID> (r);
+        q.append<T, ID> (r, conversion_);
         return q;
       }
 
@@ -632,7 +646,7 @@ namespace odb
       {
         query q (table_, column_);
         q += "<";
-        q.append<T, ID> (v);
+        q.append<T, ID> (v, conversion_);
         return q;
       }
 
@@ -649,7 +663,7 @@ namespace odb
       {
         query q (table_, column_);
         q += "<";
-        q.append<T, ID> (r);
+        q.append<T, ID> (r, conversion_);
         return q;
       }
 
@@ -717,7 +731,7 @@ namespace odb
       {
         query q (table_, column_);
         q += ">";
-        q.append<T, ID> (v);
+        q.append<T, ID> (v, conversion_);
         return q;
       }
 
@@ -734,7 +748,7 @@ namespace odb
       {
         query q (table_, column_);
         q += ">";
-        q.append<T, ID> (r);
+        q.append<T, ID> (r, conversion_);
         return q;
       }
 
@@ -802,7 +816,7 @@ namespace odb
       {
         query q (table_, column_);
         q += "<=";
-        q.append<T, ID> (v);
+        q.append<T, ID> (v, conversion_);
         return q;
       }
 
@@ -819,7 +833,7 @@ namespace odb
       {
         query q (table_, column_);
         q += "<=";
-        q.append<T, ID> (r);
+        q.append<T, ID> (r, conversion_);
         return q;
       }
 
@@ -887,7 +901,7 @@ namespace odb
       {
         query q (table_, column_);
         q += ">=";
-        q.append<T, ID> (v);
+        q.append<T, ID> (v, conversion_);
         return q;
       }
 
@@ -904,7 +918,7 @@ namespace odb
       {
         query q (table_, column_);
         q += ">=";
-        q.append<T, ID> (r);
+        q.append<T, ID> (r, conversion_);
         return q;
       }
 
@@ -1048,6 +1062,7 @@ namespace odb
     private:
       const char* table_;
       const char* column_;
+      const char* conversion_;
     };
 
     // Provide operator+() for using columns to construct native
@@ -391,10 +396,11 @@ namespace odb
     struct query_column
     {
-      // Note that we keep shalow copies of the table and column names.
+      // Note that we keep shallow copies of the table, column, and conversion
+      // expression. The latter can be NULL.
       //
-- 
cgit v1.1